/**********************************************************************
MPEG-4 Low Delay SBR conformance tool

This software module was originally developed by

Markus Schnell     (Fraunhofer GesellschaftIIS)

and edited by


in the course of development of the MPEG-2 AAC/MPEG-4 Audio standard
ISO/IEC 13818-7, 14496-1,2 and 3. This software module is an
implementation of a part of one or more MPEG-2 AAC/MPEG-4 Audio tools
as specified by the MPEG-2 AAC/MPEG-4 Audio standard. ISO/IEC gives
users of the MPEG-2 AAC/MPEG-4 Audio standards free license to this
software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-2 AAC/ MPEG-4 Audio
standards. Those intending to use this software module in hardware or
software products are advised that this use may infringe existing
patents. The original developer of this software module and his/her
company, the subsequent editors and their companies, and ISO/IEC have
no liability for use of this software module or modifications thereof
in an implementation. Copyright is not released for non MPEG-2
AAC/MPEG-4 Audio conforming products. The original developer retains
full right to use the code for his/her own purpose, assign or donate
the code to a third party and to inhibit third party from using the
code for non MPEG-2 AAC/MPEG-4 Audio conforming products. This
copyright notice must be included in all copies or derivative works.

Copyright (c) 2000.

$Id: confldsbr.c,v 1.12 2012-08-03 17:30:50 uid30883 Exp $
Decoder frame work


Source file: confldsbr.c.c


Required modules:
common.o                common module
cmdline.o               command line module
bitstream.o             bit stream module
audio.o                 audio i/o module
dec_par.o               decoder core (parametric)
dec_lpc.o               decoder core (CELP)
dec_tf.o                decoder core (T/F)

Authors:
SNL   Markus Schnell, Fraunhofer IIS <snl@iis.fraunhofer.de>

Changes:
09-mar-09   SNL   first version
**********************************************************************/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmdline.h"             /* command line module */
#include "common_m4a.h"          /* common module       */
#include "audio.h"               /* audio i/o module    */
#include "mp4au.h"               /* frame work common declarations */

#include "mp4ifc.h"
#include "streamfile.h"
#include "ct_sbrdecoder.h"
#include "ct_sbrdec.h"
#include "ldsbrdecoderConfTool.h"

#define TWIDDLEFILE   "ld-twiddle.dat"

/*
  Version info
*/
static char *PROGVER ="MPEG-4 Low Delay SBR Conformance Tool (Edition 2009)";
static char *CVSID   ="$Id: confldsbr.c,v 1.12 2012-08-03 17:30:50 uid30883 Exp $";

/* support for QMF data interface */
int qmfDataFileUsed = 0;
FILE qmfDataFileHandle;
char qmfDataFileName[] = "qmfdata.dat"; 
char *qmfDataFile = qmfDataFileName;

/* externsion payload writer support */
char payloadDataFileName[] = "mpspayload.dat"; 
extern char*  payloadDataFile;
extern int   payloadDataFileUsed;


char conf_result[256];
char conf_resultQmf[256];
int fileIndex;
/*
  command line module
*/


static int   maxLayer=1;
static int   programNr;
static char* decPara=" ";
static char* testwav;
static int   testwav_defined;
static char* bsfile;
static int   bsfile_defined;
static int   loopInFile=1;
static char* bitPath;
static char* decPath;
static int   bitPathUsed;
static int   decPathUsed;
static char* bitExt;
static int   int24flag=1;
static int   monoFilesOut;
static char* decExt;
static int   decExtUsed;
static int   mainDebugLevel=0;
static int   audioDebugLevel=0;
static int   bitDebugLevel=0;
static int   cmdDebugLevel=0;
static char* aacDebugString; /* Debug options for aac */
static int   epDebugLevel=0;
static int*  varArgIdx;
#ifdef CT_SBR
static int   HEaacProfileLevel;
#endif

/* TODO get these params somehere */
static int verbose = 0;
char *progName = "<no program name>";
/*
  debug variables
*/

#define INF_FILE_LEN 1024
static int   epFlag;
static char  infoFileName[INF_FILE_LEN];/* for EP mode only */

#ifdef DEBUGPLOT
extern int firstFrame;  /* global for debugging purposes */
#endif


static CmdLinePara paraList[] = {
  {&varArgIdx,NULL,"-b BitStream -t WavUnderTest"},
  {NULL,NULL,NULL}
};

static CmdLineSwitch switchList[] = {
  {"h",NULL,NULL,NULL,NULL,"print help"},
  {"b",&bsfile,"%s",NULL,NULL,"bit stream file name"},
  {"t",&testwav,"%s",NULL,NULL,"wav under test file name"},
  {"d",&mainDebugLevel,"%d",NULL,NULL,"main debug level (default: 0)"}, 
#if 0
  /** HACK these switches are enabled to allow debugging **/
  {"i",&testid,"%d",NULL,NULL,"testid for sequence"},
#ifdef CT_SBR
  {"HEAacProfileLevel",&HEaacProfileLevel,"%i","5",NULL,"Level of the High Efficiency AAC Profile"},
#endif
#endif
  {NULL,NULL,NULL,NULL,NULL,NULL}
};

static void MakeInfoFileName(char *bitFileName, char *infoFileName)
{
  char *writepos;
  short len;

  assert( INF_FILE_LEN > (int) strlen(bitFileName) );
  strcpy(infoFileName, bitFileName);
  len = strlen(bitFileName);
  writepos = infoFileName + len;

  if ((len >= 4) && (!strcmp(writepos - 4, ".bit") || !strcmp(writepos - 4, MP4_BIT_EXT)))
    writepos -= 4;
  sprintf(writepos,".inf");
}

static void getBsInfo ( char * bitstreaminputfile,
                        int *downsampledSbr,
                        int* samplingrate,
                        int *k0,
                        int *k1,
			int *LDMPS_active,
			int *SAOC_active
                       )
{

    HANDLE_STREAMFILE stream = NULL;
    AUDIO_SPECIFIC_CONFIG* asc;
    HANDLE_STREAMPROG prog = NULL;
    SBRDECODER ct_sbrDecoder;
    int framesize;
    unsigned char * sbrHeader;
    int i = 0;

    if(MP4Audio_ProbeFile(bitstreaminputfile)) {
        stream = StreamFileOpenRead( bitstreaminputfile, FILETYPE_AUTO );
        prog = StreamFileGetProgram( stream, /*programNr*/0 );
        asc = &(prog->decoderConfig[0].audioSpecificConfig);
	
	while (asc->specConf.eldSpecificConfig.eldExtension[i].tag != ELDEXT_TERM) {
	  if (asc->specConf.eldSpecificConfig.eldExtension[i].tag == ELDEXT_LDSAC)
	    *LDMPS_active = 1;
	  if (asc->specConf.eldSpecificConfig.eldExtension[i].tag == ELDEXT_SAOC)
	    *SAOC_active = 1;
	  i++;
	}

        /* some sanity checks */
        if (asc->audioDecoderType.value != 39)
            CommonExit(1,"AOT is not MPEG-4 ER AAC ELD");

        framesize = asc->specConf.eldSpecificConfig.frameLengthFlag.value ? 480 : 512;
        *downsampledSbr = !asc->specConf.eldSpecificConfig.ldSbrSamplingRate.value;
        *k0 = 0; *k1 = 0;

        if (asc->specConf.eldSpecificConfig.ldSbrPresentFlag.value == 1)
        {
            /* CommonExit(1,"no LD-SBR active"); */
          sbrHeader =  asc->specConf.eldSpecificConfig.sbrHeaderData[0];

          *samplingrate = asc->samplingFrequency.value;

          ct_sbrDecoder = openSBR(asc->samplingFrequency.value,
                                  *downsampledSbr,
                                  framesize
          #ifdef AAC_ELD
                                  ,1 /*ldsbr*/
                                  ,sbrHeader
          #endif
                                 );
          *k0 = getStartBand(0);
          *k1 = getStopBand(0);
          closeSBR ( ct_sbrDecoder );
        }
        StreamFileClose(stream);
    }
    else{
        CommonExit(1, "error open bit stream %s", bitstreaminputfile);
    }
}

static void writeTwiddles (int bDownSampledSbr, Twiddle twiddle) {
  FILE *twiddlefile;

  if ((twiddlefile = fopen(TWIDDLEFILE, "rb+")) || (twiddlefile = fopen(TWIDDLEFILE, "wb+"))) {
    if (bDownSampledSbr) {
      fseek(twiddlefile, 2 * sizeof(double), SEEK_SET);
    }
    fwrite(&twiddle.fi,   1, sizeof(double), twiddlefile);
    fwrite(&twiddle.beta, 1, sizeof(double), twiddlefile);
    fclose(twiddlefile);
    if (verbose)
      printf("%s: %s written\n", progName, TWIDDLEFILE);
  } else {
    CommonWarning("couldn't open or create %s. Maybe permission denied?", TWIDDLEFILE);
  }
}

static void estimateTwiddles (char *audioinput, int *bDownSampledSbr, int *samplingrate, int *k0, int *k1, int useMCextension)
{
    float **testBuf;

    AudioFile *testFile;
    int testChannel, framesize, ch, chan2Analyze;
    float testSample;
    long testNumSample, tmpSample;

    /* due to the non-linear phase filter bank of the ELD we cannot rely on correct twiddling in the
     * first and last SBR bands. Therefore, skip some: */
    int k0_eld = *k0 + 2;
    int k1_eld = *k1 - 2;

    Twiddle twiddle;

    testFile = AudioOpenRead(audioinput,&testChannel,&testSample,&testNumSample);

    chan2Analyze = testChannel;
    if (useMCextension) 
      chan2Analyze = 1;

    if ( testSample != *samplingrate * (2 - *bDownSampledSbr) )
    {
      CommonExit(1, "sampling rates of input files differ!\n");
    }

    framesize = 16384 * (2 - *bDownSampledSbr);
    if ((testBuf=(float**)malloc(testChannel*sizeof(float*)))==NULL)
          CommonExit(1,"memory allocation error");
        for (ch=0; ch<testChannel; ch++)
          if ((testBuf[ch]=(float*)malloc(framesize*testChannel*sizeof(float)))==NULL)
            CommonExit(1,"memory allocation error");
    tmpSample = AudioReadData(testFile, testBuf, framesize*testChannel);

    AudioClose(testFile);

    twiddle = identifyQmfBank( (float**)testBuf[0],
                               chan2Analyze,
                               *bDownSampledSbr,
                               k0_eld,
                               k1_eld );
    if (verbose)
      printf("%s: twiddles estimated\n", progName);

    writeTwiddles(*bDownSampledSbr, twiddle);

    free(testBuf);
}

static void setTwiddles (int bDownSampledSbr) {

  FILE *twiddlefile;
  extern Twiddle* twiddles;

#ifdef LOW_POWER_SBR
  twiddles = NULL;
  return;
#endif
  if( (twiddlefile = fopen(TWIDDLEFILE, "rb")) ) {
    double fi, beta;
    unsigned int count;
    count  = fread(&fi,   1, sizeof(double), twiddlefile);
    count += fread(&beta, 1, sizeof(double), twiddlefile);
    if(bDownSampledSbr) {
      count += fread(&fi,   1, sizeof(double), twiddlefile);
      count += fread(&beta, 1, sizeof(double), twiddlefile);
    }

    if(count == 2 * sizeof(double) * (bDownSampledSbr + 1)) {
      if ((twiddles=(Twiddle*)malloc(sizeof(Twiddle)))==NULL)
        CommonExit(1,"memory allocation error");

      twiddles->fi   = fi;
      twiddles->beta = beta;

      if (verbose)
	printf("%s: using fi = %f, beta = %f\n", progName, fi, beta);
    }
    else {
      twiddles = NULL;
      CommonWarning("missing twiddles for downsampled SBR -> ignoring twiddle-file\n");
    }

    fclose(twiddlefile);
  }
  else {
    twiddles = NULL;
    if (fileIndex != 5)
      CommonWarning("ignoring twiddle-file\n");
  }
}

static void runTest (int testid, float **refBuf, float**testBuf, int nChannel, int frameSize) {

  int offset = 0;
  switch (testid) {
  case 1: /* ELD */
    fileIndex = 1;
    calculateDiffandRms(refBuf, testBuf, nChannel, frameSize, offset);
    break;
  case 2: /* ELD + LDMPS */    
    fileIndex = 1;    
    calculateDiffandRms(refBuf, testBuf, 1, frameSize, 1024+512);   
    break;
  case 3: /* ELD + LDMPS */    
    break;
  default:
    CommonExit (1,"test id not found");
  }

}


/*
  main
*/ 

int main (int argc, char *argv[])
{
  int testid = 1;  /* ELD is default; 2 = ELD + LDMPS */
  int  result;
  char bitFileName[STRLEN];
  char *decTmpName = "tmpConfDecFile";
  char decFileName[STRLEN];
  char *decFormat;
  extern Twiddle *twiddles;
  int useTwiddles=1;
  float **refBuf, **testBuf;

  AudioFile *referenceFile;
  AudioFile *testFile;

  int bDownSampledSbr;
  int sbrStart, sbrStop;
  int samplingrate;

  int LDMPS_active = 0;
  int SAOC_active  = 0;
  /* evaluate command line  */

  CmdLineInit(0);

  result = CmdLineEval(argc,argv,paraList,switchList,1,&progName);

  if (argc == 1) result=1;

  if (result) {
    if (result==1) {
      printf("%s: %s\n",progName,PROGVER);
      CmdLineHelp(progName,paraList,switchList,stdout);
      exit (1);
    }
    else
      CommonExit(1,"command line error (\"-h\" for help)");
  }

  CmdLineInit(cmdDebugLevel);

  /* set everything quiet except maindebuglevel is set up */ 
  /* for mp4audec debug level is decreased by one */ 
  verbose = (mainDebugLevel > 0); 
  mainDebugLevel -= 1; 
  mainDebugLevel = (mainDebugLevel < 0) ? 0 : mainDebugLevel; 
  SetDebugLevel (mainDebugLevel) ; 
 
  if (1) 
    printf("%s: %s\n",progName,PROGVER); 
  if (mainDebugLevel >= 1) { 
    printf("CVS Id: %s\n",CVSID); 
  } 

  /* calc variable default values */
  if (!bitPathUsed)
    bitPath = getenv(MP4_BIT_PATH_ENV);
  if (!decPathUsed)
    decPath = getenv(MP4_DEC_PATH_ENV);
  decExt = ".wav";

  if (bsfile == NULL)
    CommonExit(1,"no bit stream file given. specify with option -b");
  if (testwav == NULL)
    CommonExit(1,"no test wav file given. specify with option -t");

  {
    /* compose file names */
    if (ComposeFileName(bsfile,0,bitPath,bitExt,bitFileName, STRLEN))
      CommonExit(1,"composed file name too long");
    if ( epFlag )
      MakeInfoFileName(bitFileName, infoFileName);

      if (ComposeFileName(decTmpName,1,decPath,decExt,decFileName,STRLEN))
        CommonExit(1,"composed file name too long");

    /* extract file format (extension) */
    if (strcmp(decFileName,"-") == 0)
      decFormat = decExt;
    else {
      decFormat = decFileName+strlen(decFileName);
      do
        decFormat--;
      while (decFormat>decFileName && *decFormat!='.');
    }

    /* decode file */
    if (verbose)
      printf("%s: decoding %s -> %s\n",progName,bitFileName,decFileName);

    getBsInfo (bitFileName,
               &bDownSampledSbr,
               &samplingrate,
               &sbrStart,
               &sbrStop,
	       &LDMPS_active,
	       &SAOC_active
	       );

    if (LDMPS_active || SAOC_active) {
      payloadDataFile = payloadDataFileName;
      payloadDataFileUsed = 1;
      qmfDataFileUsed = 1;
    }

    if(     strstr(bsfile,"sbr_twi_") 
        ||  strstr(bsfile,"er_eld_ldmps_ldqmf_") 
      )
    {
      estimateTwiddles(testwav,
                       &bDownSampledSbr,
                       &samplingrate,
                       &sbrStart,
                       &sbrStop,
		       LDMPS_active);
    } 

    setTwiddles (bDownSampledSbr);

    if (MP4Audio_ProbeFile (bitFileName))
    {
        MP4Audio_DecodeFile (bitFileName,
                             decFileName,
                             decFormat,
                             int24flag,
                             monoFilesOut,
                             maxLayer,
                             decPara,
                             mainDebugLevel,
                             audioDebugLevel,
                             epDebugLevel,
                             bitDebugLevel,
                             aacDebugString,
#ifdef CT_SBR
                             HEaacProfileLevel,
#endif
                             programNr
                             );
    }
    else{
      CommonExit(1,"error decoding audio file %s",bitFileName);
    }

  } /* end decode */

  {
    /* file decode so open testfile and decoded file */
    int refChannel, testChannel, framesize, ch;
    float refSample, testSample;
    long refNumSample, testNumSample, totNumSample, curNumSample, tmpSample;

    if (LDMPS_active) {
      if (strstr(bsfile,"er_eld_ldmps_ldqmf_"))
	testid = 2;
      else
	testid = 3;
    }

    referenceFile = AudioOpenRead(decFileName,&refChannel,&refSample,&refNumSample);
    if (referenceFile==NULL)
      CommonExit(1,"error opening reference file %s ",decFileName);

    testFile = AudioOpenRead(testwav,&testChannel,&testSample,&testNumSample);
    if (testFile==NULL)
      CommonExit(1,"error opening test wav file %s ",testwav);

    /* some sanity checks */
    if (refSample != testSample)
      CommonExit(1,"sample rate differs");

    if (!LDMPS_active) {
      if (refChannel != testChannel)
	CommonExit(1,"number of channels differ");
    } else {
      if (!(refChannel==1 /* && testChannel==2 */) )
	CommonExit(1,"Channels configuration differs from: reference 1ch, test 2ch");
    }

    if (refNumSample != testNumSample) {
      CommonWarning("number of audio samples differ -> take the minimum");
      totNumSample = MIN(refNumSample, testNumSample);
    } else
      totNumSample = refNumSample;

      /* loop for read and compare */
    curNumSample = 0;
    framesize = totNumSample/refChannel;   /* read the complete file at once */

    if ((refBuf=(float**)malloc(refChannel*sizeof(float*)))==NULL)
      CommonExit(1,"memory allocation error");
    for (ch=0; ch<refChannel; ch++)
      if ((refBuf[ch]=(float*)malloc(framesize*refChannel*sizeof(float)))==NULL)
        CommonExit(1,"memory allocation error");

    if ((testBuf=(float**)malloc(testChannel*sizeof(float*)))==NULL)
      CommonExit(1,"memory allocation error");
    for (ch=0; ch<testChannel; ch++)
      if ((testBuf[ch]=(float*)malloc(framesize*testChannel*sizeof(float)))==NULL)
        CommonExit(1,"memory allocation error");

    while (curNumSample < totNumSample) {
      tmpSample = AudioReadData(referenceFile, refBuf, framesize*refChannel);
      tmpSample = AudioReadData(testFile, testBuf, framesize*refChannel);
      curNumSample += tmpSample;
      
      runTest (testid, refBuf, testBuf, refChannel, framesize);
    }

    if (refChannel>1) {
      for (ch=0; ch<refChannel; ch++) {
        free(refBuf[ch]);
        free(testBuf[ch]);
      }
    }
    free(refBuf);
    free(testBuf);

    AudioClose(referenceFile);
    AudioClose(testFile);

    if (twiddles != NULL) {
        free (twiddles);
        twiddles = NULL;
    }

  }
  CmdLineEvalFree(paraList);


  if (testid != 3) {
    printf("%s\n", conf_resultQmf);
    printf("%s\n", conf_result);
    printf("\n");
    CommonWarning("RMS/maxDiff thresholds for the various test sequences have not been adjusted yet"); 
    CommonWarning("-> PASS or FAIL decision might not be correct"); 
  } else {
    printf("Further decoding necessary for final result\n");
  }

  if (LDMPS_active) {
    printf("Please pass on the files mpspayload.dat, qmfdata.dat and ld-twiddle.dat (if available)");
    printf(" to the MPEG-D LDSAC decoder \n");
  }


  return 0;
}

/* end of mp4dec.c */

