/****************************************************************************

SC 29 Software Copyright Licencing Disclaimer:

This software module was originally developed by
  Fraunhofer GhG
  Dolby

and edited by
  -

in the course of development of the ISO/IEC 13818-7 and ISO/IEC 14496-3
standards for reference purposes and its performance may not have been
optimized. This software module is an implementation of one or more tools as
specified by the ISO/IEC 13818-7 and ISO/IEC 14496-3 standards.
ISO/IEC gives users free license to this software module or modifications
thereof for use in products claiming conformance to audiovisual and
image-coding related ITU Recommendations and/or ISO/IEC International
Standards. ISO/IEC gives users the same free license to this software module or
modifications thereof for research purposes and further ISO/IEC standardisation.
Those intending to use this software module in products are advised that its
use may infringe existing patents. ISO/IEC have no liability for use of this
software module or modifications thereof. Copyright is not released for
products that do not conform to audiovisual and image-coding related ITU
Recommendations and/or ISO/IEC International Standards.
The original developer retains full right to modify and use the code for its
own purpose, assign or donate the code to a third party and to inhibit third
parties from using the code for products that do not conform to audiovisual and
image-coding related ITU Recommendations and/or ISO/IEC International Standards.
This copyright notice must be included in all copies or derivative works.
Copyright (c) ISO/IEC 2002.



*******************************************************************************/
/*!
  \file
  \brief  SBR decoder frontend $Revision: 1.4 $
*/


#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "ldsbrdecoderConfTool.h"


int fileIndex = -1;
const float conCriteria[][2] = {
  {0.0f, 0.0f}, /* sbr_qmf */
  {0.0f, 0.0f}, /* sbr_e   */
  {0.0f, 0.0f}, /* sbr_gh  */
  {0.0f, 0.0f}, /* sbr_i   */
  {0.0f, 0.0f}, /* sbr_s   */
  {0.0f, 0.0f}  /* sbr_twi */
#ifdef PARAMETRICSTEREO
  ,
  {0.0f,   0.0f},
  {0.0f,   0.0f}
  #ifndef NO_PS_CONF_DCOR2
  /* these conformance criteria have been proposed in 14496-4:2004/Amd.11:2006/DCOR 2
   * balloting for this hasn't closed at the time of this writing, so they are not official.
   * Hence they can be removed from the build by defining NO_PS_CONF_DCOR2 in this file and in decoderConfTool.c
   */
  ,{0.0f,   0.0f},
  {0.0f,   0.0f},
  {0.0f,   0.0f},
  {0.0f,   0.0f},
  {0.0f,   0.0f}
  #endif
#endif
};



#ifdef PARAMETRICSTEREO
  extern int doPsConformance;
#endif

void
calculateDiffandRms(float** timeData,
                    float** timeDataRef,
                    int nChannels,
                    int frameSize,
		    int offset)
{
  int channelIndex, i;
  static double maxDiff = 0;
  int maxPos = 0;
  static double totalDiff = 0;
  static int count = 0;
  double diff;
  extern char conf_result[256];
  int tmp;

  for (channelIndex = 0 ; channelIndex < nChannels ; channelIndex++){
      for(i= offset; i < frameSize; i++){
        diff = fabs(timeDataRef[channelIndex][i ] - timeData[channelIndex][i]);
        totalDiff += pow(timeDataRef[channelIndex][i ] - timeData[channelIndex][i],2);

        count++;
        if(diff > maxDiff){
          maxPos = i;
          maxDiff = diff;
        }
      }
  }

  if (fileIndex == 0 || fileIndex == 5) {
      tmp = sprintf(conf_result, "%s  QMF-TEST: Max Diff: %.3f (threshold %.3f) RMS Diff: %.3f  (threshold %.3f)", GREP_STRING, maxDiff,conCriteria[fileIndex][0], sqrt(totalDiff/count), conCriteria[fileIndex][1]);
  }
  else if (fileIndex >=1 && fileIndex <=4) {
      tmp = sprintf(conf_result, "%s  SBR-TEST: Max Diff: %.3f (threshold %.3f) RMS Diff: %.3f  (threshold %.3f)", GREP_STRING, maxDiff,conCriteria[fileIndex][0], sqrt(totalDiff/count), conCriteria[fileIndex][1]);
  }
#ifdef PARAMETRICSTEREO
  else if (fileIndex >=6 && fileIndex <=
#ifndef NO_PS_CONF_DCOR2
           12
#else
           7
#endif
           ) {
    tmp = sprintf(conf_result, "%s  SBR+PS-TEST: Max Diff: %.3f (threshold %.3f) RMS Diff: %.3f  (threshold %.3f)", GREP_STRING, maxDiff,conCriteria[fileIndex][0], sqrt(totalDiff/count), conCriteria[fileIndex][1]);
  }
#endif
  else {
    tmp = sprintf(conf_result, "%s  SBR-TEST: Max Diff: %.3f RMS Diff: %.3f", GREP_STRING, maxDiff, sqrt(totalDiff/count));
  }
  if (fileIndex != -1) {
    if (fileIndex != 5) {
#if 1 /* reset this when conformance criteria for maxDiff and RMS are defined! */
    	tmp += sprintf(conf_result + tmp, ", -> Conformance NOT_DEFINED");
#else
      if (maxDiff < conCriteria[fileIndex][0] && sqrt(totalDiff/count) < conCriteria[fileIndex][1])
          tmp += sprintf(conf_result + tmp, ", -> Conformance PASSED");
      else
          tmp += sprintf(conf_result + tmp, ", -> Conformance FAILED");
#endif

    }
#if !defined LOW_POWER_SBR
    else {
      if (maxDiff < conCriteria[fileIndex][0] && sqrt(totalDiff/count) < conCriteria[fileIndex][1])
          tmp += sprintf(conf_result + tmp, ", -> Twiddle estimation succeded");
      else
        tmp += sprintf(conf_result + tmp, ", -> Twiddle estimation might not have succeeded; later conformance tests might fail for this reason.");
    }
#endif
  }
  memset(conf_result + tmp, 0, sizeof(conf_result) - tmp);
}



void
calculateDiffandRmsQmf(float** timeData,
                       float** timeDataRef,
                       int nChannels,
                       int frameSize)
{
  int channelIndex, i;
  static double maxDiffQmf = 0;
  static double totalDiffQmf = 0;
  static int countQmf = 0;
  int tmp;

  extern char conf_resultQmf[256];

  for (channelIndex = 0 ; channelIndex < nChannels ; channelIndex++)
  {
      for(i= 0; i < frameSize; i++){
        double diff = fabs(timeDataRef[channelIndex][i] - timeData[channelIndex][i]);
        totalDiffQmf += pow((double)timeDataRef[channelIndex][i] - (double)timeData[channelIndex][i],2);

        countQmf++;
        if(diff > maxDiffQmf){
          maxDiffQmf = diff;
        }
      }
  }

  if (fileIndex != -1) {
    tmp = sprintf(conf_resultQmf, "%s  (Informative) QMF-TEST: Max Diff: %.3f (%.3f) RMS Diff: %.3f  (%.3f)", GREP_STRING, maxDiffQmf,conCriteria[fileIndex][0], sqrt(totalDiffQmf/countQmf), conCriteria[fileIndex][1]);
  }
  else {
      tmp = sprintf(conf_resultQmf, "%s  (Informative) QMF-TEST: Max Diff: %.3f RMS Diff: %.3f", GREP_STRING, maxDiffQmf, sqrt(totalDiffQmf/countQmf));
  }

  memset(conf_resultQmf + tmp, 0, sizeof(conf_resultQmf) - tmp);

}

Twiddle
identifyQmfBank( const float *inData,
                 int noChannels,
                 int bDownSampledSbr,
                 int k0,              /* first QMF channel in analysis, zero-index based            */
                 int k1 )             /* last QMF channel in analysis (exclusive), zero-index based */

{
  Twiddle twiddle;
  int n, k, nch, K, L;
  int F = 32768;        /* analysis length        */
  int M =    64;        /* QMF bank size          */

  double s0, s1, a, b;
  double Hr[64] = { 0 };
  double Hi[64] = { 0 };
  double Ph[64] = { 0 };

  if(bDownSampledSbr) {
    F /= 2;
    M /= 2;
    k1 = MIN(32, k1);
  }

  /* compute DFT with frequency offset */
  for(k = k0; k < k1; k++) {
    for(n = 0; n < F; n++) {
      Hr[k] += inData[n * noChannels] * cos(2 * PI / (2*M) * (k+0.5) * n);
      Hi[k] -= inData[n * noChannels] * sin(2 * PI / (2*M) * (k+0.5) * n);
    }
  }

  /* compute unwrapped phase */
  for(k = k0; k < k1; k++) {

    Ph[k] = atan2( Hi[k], Hr[k] ) / PI;

    if(Ph[k] > Ph[k-1])
      while(Ph[k] - Ph[k-1] > 1)
        Ph[k] -= 2;
    else
      while(Ph[k-1] - Ph[k] > 1)
        Ph[k] += 2;
  }

  /* linear curve fitting */
  nch = k1 - k0;                                               /* number of channels in analysis */

  /* compute factors */
  K = ( k1 * (k1-1) * (2*k1-1) - k0 * (k0-1) * (2*k0-1) ) / 6; /* Sum(k^2) from k0 to k1-1. */
  L = ( k1 * (k1-1) - k0 * (k0-1) ) / 2;                       /* Sum(k)   from k0 to k1-1. */

  /* compute synthesis twiddle f(k) = a * k + b */
  s0 = s1 = 0;
  for(k = k0; k < k1; k++) {
    s0 += k * Ph[k];
    s1 += Ph[k];
  }

  a = ( nch * s0 - L * s1 ) / (K * nch - L * L);
  b = ( s0 - K * a ) / L;

  /* rewrite phase j*pi*(a*k+b) as j*pi/L*((k+0.5)*fi+beta). */
  twiddle.fi   = M * a;
  twiddle.beta = M * b - M/2 * a;

  /* Round fi in 1/M steps */
  if(twiddle.fi >= 0)
    twiddle.fi = (int) (M * twiddle.fi   + 0.5);
  else
    twiddle.fi = (int) (M * twiddle.fi   - 0.5);

  twiddle.fi  *= (1.0 / M);

  /* Round beta in steps of one */
  if(twiddle.beta >= 0)
    twiddle.beta  = (int) (twiddle.beta + 0.5);
  else
    twiddle.beta  = (int) (twiddle.beta - 0.5);

  /* align sign of beta with fi (optional) */
  if(twiddle.fi > 0) {
    while(twiddle.beta < 0)
      twiddle.beta += 2 * M;
  } else if(twiddle.fi < 0) {
    while(twiddle.beta > 0)
      twiddle.beta -= 2 * M;
  }

  return twiddle;
}

